# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
All functions for converting weather data into energy system model data.
"""

import xarray as xr
import numpy as np
import pandas as pd
import geopandas as gpd
import datetime as dt
from operator import itemgetter
from pathlib import Path
from dask.diagnostics import ProgressBar
from scipy.sparse import csr_matrix

import logging
logger = logging.getLogger(__name__)

from .aggregate import aggregate_matrix
from .gis import spdiag

from .pv.solar_position import SolarPosition
from .pv.irradiation import TiltedIrradiation
from .pv.solar_panel_model import SolarPanelModel
from .pv.orientation import get_orientation, SurfaceOrientation

from . import hydro as hydrom
from . import wind as windm

from .resource import (get_windturbineconfig, get_solarpanelconfig,
                       windturbine_smooth)


def convert_and_aggregate(cutout, convert_func, matrix=None,
                          index=None, layout=None, shapes=None,
                          shapes_crs=4326, per_unit=False,
                          return_capacity=False, capacity_factor=False,
                          show_progress=True, dask_kwargs={}, **convert_kwds):
    """
    Convert and aggregate a weather-based renewable generation time-series.

    NOTE: Not meant to be used by the user him or herself. Rather it is a
    gateway function that is called by all the individual time-series
    generation functions like pv and wind. Thus, all its parameters are also
    available from these.

    Parameters
    -----------
    matrix : N x S - xr.DataArray or sp.sparse.csr_matrix or None
        If given, it is used to aggregate the grid cells to buses.
        N is the number of buses, S the number of spatial coordinates, in the
        order of `cutout.grid`.
    index : pd.Index
        Index of Buses.
    layout : X x Y - xr.DataArray
        The capacity to be build in each of the `grid_cells`.
    shapes : list or pd.Series of shapely.geometry.Polygon
        If given, matrix is constructed as indicatormatrix of the polygons, its
        index determines the bus index on the time-series.
    shapes_crs : pyproj.CRS or compatible
        If different to the map crs of the cutout, the shapes are
        transformed to match cutout.crs (defaults to EPSG:4326).
    per_unit : boolean
        Returns the time-series in per-unit units, instead of in MW (defaults
        to False).
    return_capacity : boolean
        Additionally returns the installed capacity at each bus corresponding
        to `layout` (defaults to False).
    capacity_factor : boolean
        If True, the static capacity factor of the chosen resource for each
        grid cell is computed.
    show_progress : boolean, default True
        Whether to show a progress bar.
    dask_kwargs : dict, default {}
        Dict with keyword arguments passed to `dask.compute`.

    Other Parameters
    -----------------
    convert_func : Function
        Callback like convert_wind, convert_pv


    Returns
    -------
    resource : xr.DataArray
        Time-series of renewable generation aggregated to buses, if
        `matrix` or equivalents are provided else the total sum of
        generated energy.
    units : xr.DataArray (optional)
        The installed units per bus in MW corresponding to `layout`
        (only if `return_capacity` is True).

    """

    func_name = convert_func.__name__.replace('convert_', '')
    logger.info(f"Convert and aggregate '{func_name}'.")
    da = convert_func(cutout.data, **convert_kwds)

    no_args = all(v is None for v in [layout, shapes, matrix])

    if no_args:
        if per_unit or return_capacity:
            raise ValueError("One of `matrix`, `shapes` and `layout` must be "
                             "given for `per_unit` or `return_capacity`")
        if capacity_factor:
            return maybe_progressbar(da.mean('time'), show_progress)
        else:
            return maybe_progressbar(da.sum('time'), show_progress)


    if shapes is not None:
        geoseries_like = (pd.Series, gpd.GeoDataFrame, gpd.GeoSeries)
        if isinstance(shapes, geoseries_like) and index is None:
            index = shapes.index
        matrix = cutout.indicatormatrix(shapes, shapes_crs)


    if layout is not None:
        assert isinstance(layout, xr.DataArray)
        layout = layout.reindex_like(cutout.data).stack(spatial=['y', 'x'])
        if matrix is None:
            matrix = layout.expand_dims('new')
        else:
            matrix = csr_matrix(matrix) * spdiag(layout)

    index = pd.RangeIndex(matrix.shape[0]) if index is None else index
    matrix = csr_matrix(matrix)
    results = aggregate_matrix(da, matrix=matrix, index=index)


    if per_unit or return_capacity:
        caps = matrix.sum(-1)
        capacity = xr.DataArray(np.asarray(caps).flatten(), [index])
        capacity.attrs['units'] ='MW'

    if per_unit:
        results = (results / capacity.where(capacity!=0)).fillna(0.)
        results.attrs['units'] = 'p.u.'
    else:
        results.attrs['units'] = 'MW'


    if return_capacity:
        return maybe_progressbar(results, show_progress, **dask_kwargs), capacity
    else:
        return maybe_progressbar(results, show_progress, **dask_kwargs)


def maybe_progressbar(ds, show_progress, **kwargs):
    """Load a xr.dataset with dask arrays either with or without progressbar."""
    if show_progress:
        with ProgressBar(minimum=2):
            ds.load(**kwargs)
    else:
        ds.load(**kwargs)
    return ds


# temperature
def convert_temperature(ds):
    """Return outside temperature (useful for e.g. heat pump T-dependent
    coefficient of performance).
    """

    # Temperature is in Kelvin
    return ds['temperature'] - 273.15


def temperature(cutout, **params):
    return cutout.convert_and_aggregate(
        convert_func=convert_temperature, **params)


# soil temperature
def convert_soil_temperature(ds):
    """Return soil temperature (useful for e.g. heat pump T-dependent
    coefficient of performance).
    """

    # Temperature is in Kelvin

    # There are nans where there is sea; by setting them
    # to zero we guarantee they do not contribute when multiplied
    # by matrix in atlite/aggregate.py
    return (ds['soil temperature'] - 273.15).fillna(0.)


def soil_temperature(cutout, **params):
    return cutout.convert_and_aggregate(
        convert_func=convert_soil_temperature, **params)


# heat demand
def convert_heat_demand(ds, threshold, a, constant, hour_shift):
    # Temperature is in Kelvin; take daily average
    T = ds['temperature']
    T = T.assign_coords(time=(T.coords['time'] +
                              np.timedelta64(dt.timedelta(hours=hour_shift))))

    T = T.resample(time="1D").mean(dim='time')
    threshold += 273.15
    heat_demand = a * (threshold - T)

    heat_demand = heat_demand.clip(min=0.)

    return (constant + heat_demand).rename('heat_demand')


def heat_demand(
        cutout,
        threshold=15.,
        a=1.,
        constant=0.,
        hour_shift=0.,
        **params):
    """
    Convert outside temperature into daily heat demand using the
    degree-day approximation.

    Since "daily average temperature" means different things in
    different time zones and since xarray coordinates do not handle
    time zones gracefully like pd.DateTimeIndex, you can provide an
    hour_shift to redefine when the day starts.

    E.g. for Moscow in winter, hour_shift = 4, for New York in winter,
    hour_shift = -5

    This time shift applies across the entire spatial scope of ds for
    all times. More fine-grained control will be built in a some
    point, i.e. space- and time-dependent time zones.

    WARNING: Because the original data is provided every month, at the
    month boundaries there is untidiness if you use a time shift. The
    resulting xarray will have duplicates in the index for the parts
    of the day in each month at the boundary. You will have to
    re-average these based on the number of hours in each month for
    the duplicated day.

    Parameters
    ----------
    threshold : float
        Outside temperature in degrees Celsius above which there is no
        heat demand.
    a : float
        Linear factor relating heat demand to outside temperature.
    constant : float
        Constant part of heat demand that does not depend on outside
        temperature (e.g. due to water heating).
    hour_shift : float
        Time shift relative to UTC for taking daily average

    Note
    ----
    You can also specify all of the general conversion arguments
    documented in the `convert_and_aggregate` function.
    """

    return cutout.convert_and_aggregate(convert_func=convert_heat_demand,
                                        threshold=threshold, a=a,
                                        constant=constant,
                                        hour_shift=hour_shift,
                                        **params)


# solar thermal collectors
def convert_solar_thermal(ds, orientation, trigon_model, clearsky_model,
                          c0, c1, t_store):
    # convert storage temperature to Kelvin in line with reanalysis data
    t_store += 273.15

    # Downward shortwave radiation flux is in W/m^2
    # http://rda.ucar.edu/datasets/ds094.0/#metadata/detailed.html?_do=y
    solar_position = SolarPosition(ds)
    surface_orientation = SurfaceOrientation(ds, solar_position, orientation)
    irradiation = TiltedIrradiation(ds, solar_position, surface_orientation,
                                    trigon_model, clearsky_model)

    # overall efficiency; can be negative, so need to remove negative values
    # below
    eta = c0 - c1 * ((t_store - ds['temperature']) /
                     irradiation.where(irradiation!=0)).fillna(0)

    output = irradiation * eta

    return output.where(output > 0., 0.)


def solar_thermal(cutout, orientation={'slope': 45., 'azimuth': 180.},
                  trigon_model="simple",
                  clearsky_model="simple",
                  c0=0.8, c1=3., t_store=80.,
                  **params):
    """
    Convert downward short-wave radiation flux and outside temperature
    into time series for solar thermal collectors.

    Mathematical model and defaults for c0, c1 based on model in [1].

    Parameters
    ----------
    cutout : cutout
    orientation : dict or str or function
        Panel orientation with slope and azimuth (units of degrees), or
        'latitude_optimal'.
    trigon_model : str
        Type of trigonometry model
    clearsky_model : str or None
        Type of clearsky model for diffuse irradiation. Either
        'simple' or 'enhanced'.
    c0, c1 : float
        Parameters for model in [1] (defaults to 0.8 and 3., respectively)
    t_store : float
        Store temperature in degree Celsius

    Note
    ----
    You can also specify all of the general conversion arguments
    documented in the `convert_and_aggregate` function.

    References
    ----------
    [1] Henning and Palzer, Renewable and Sustainable Energy Reviews 30
    (2014) 1003-1018
    """

    if not callable(orientation):
        orientation = get_orientation(orientation)

    return cutout.convert_and_aggregate(convert_func=convert_solar_thermal,
                                        orientation=orientation,
                                        trigon_model=trigon_model,
                                        clearsky_model=clearsky_model,
                                        c0=c0, c1=c1, t_store=t_store,
                                        **params)


# wind
def convert_wind(ds, turbine):
    """Convert wind speeds for turbine to wind energy generation."""

    V, POW, hub_height, P = itemgetter('V', 'POW', 'hub_height', 'P')(turbine)

    wnd_hub = windm.extrapolate_wind_speed(ds, to_height=hub_height)

    def _interpolate(da):
        return np.interp(da, V, POW / P)

    return xr.apply_ufunc(_interpolate, wnd_hub,
                          input_core_dims=[[]],
                          output_core_dims=[[]],
                          output_dtypes=[wnd_hub.dtype],
                          dask="parallelized")


def wind(cutout, turbine, smooth=False, **params):
    """
    Generate wind generation time-series

    Extrapolates 10m wind speed with monthly surface roughness to hub
    height and evaluates the power curve.

    Parameters
    ----------
    turbine : str or dict
        A turbineconfig dictionary with the keys 'hub_height' for the
        hub height and 'V', 'POW' defining the power curve.
        Alternatively a str refering to a local or remote turbine configuration
        as accepted by atlite.resource.get_windturbineconfig().
    smooth : bool or dict
        If True smooth power curve with a gaussian kernel as
        determined for the Danish wind fleet to Delta_v = 1.27 and
        sigma = 2.29. A dict allows to tune these values.

    Note
    ----
    You can also specify all of the general conversion arguments
    documented in the `convert_and_aggregate` function.

    References
    ----------
    [1] Andresen G B, Søndergaard A A and Greiner M 2015 Energy 93, Part 1
    1074 – 1088. doi:10.1016/j.energy.2015.09.071
    """

    if isinstance(turbine, (str, Path)):
        turbine = get_windturbineconfig(turbine)

    if smooth:
        turbine = windturbine_smooth(turbine, params=smooth)

    return cutout.convert_and_aggregate(
        convert_func=convert_wind, turbine=turbine, **params)


# solar PV
def convert_pv(
        ds,
        panel,
        orientation,
        trigon_model='simple',
        clearsky_model='simple'):
    solar_position = SolarPosition(ds)
    surface_orientation = SurfaceOrientation(ds, solar_position, orientation)
    irradiation = TiltedIrradiation(ds, solar_position, surface_orientation,
                                    trigon_model=trigon_model,
                                    clearsky_model=clearsky_model)
    solar_panel = SolarPanelModel(ds, irradiation, panel)
    return solar_panel


def pv(cutout, panel, orientation, clearsky_model=None, **params):
    """
    Convert downward-shortwave, upward-shortwave radiation flux and
    ambient temperature into a pv generation time-series.

    Parameters
    ----------
    panel : str or dict
        Panel config dictionary with the parameters for the electrical
        model in [3]. Alternatively, name of yaml file stored in
        atlite.config.solarpanel_dir.
    orientation : str, dict or callback
        Panel orientation can be chosen from either
        'latitude_optimal', a constant orientation {'slope': 0.0,
        'azimuth': 0.0} or a callback function with the same signature
        as the callbacks generated by the
        'atlite.pv.orientation.make_*' functions.
    clearsky_model : str or None
        Either the 'simple' or the 'enhanced' Reindl clearsky
        model. The default choice of None will choose dependending on
        data availability, since the 'enhanced' model also
        incorporates ambient air temperature and relative humidity.

    Returns
    -------
    pv : xr.DataArray
        Time-series or capacity factors based on additional general
        conversion arguments.

    Note
    ----
    You can also specify all of the general conversion arguments
    documented in the `convert_and_aggregate` function.

    References
    ----------
    [1] Soteris A. Kalogirou. Solar Energy Engineering: Processes and Systems,
    pages 49–117,469–516. Academic Press, 2009. ISBN 0123745012.

    [2] D.T. Reindl, W.A. Beckman, and J.A. Duffie. Diffuse fraction correla-
    tions. Solar Energy, 45(1):1 – 7, 1990.

    [3] Hans Georg Beyer, Gerd Heilscher and Stefan Bofinger. A Robust Model
    for the MPP Performance of Different Types of PV-Modules Applied for
    the Performance Check of Grid Connected Systems, Freiburg, June 2004.
    Eurosun (ISES Europe Solar Congress).
    """

    if isinstance(panel, (str, Path)):
        panel = get_solarpanelconfig(panel)
    if not callable(orientation):
        orientation = get_orientation(orientation)

    return cutout.convert_and_aggregate(convert_func=convert_pv,
                                        panel=panel, orientation=orientation,
                                        clearsky_model=clearsky_model,
                                        **params)

# hydro
def convert_runoff(ds, weight_with_height=True):
    runoff = ds['runoff']

    if weight_with_height:
        runoff = runoff * ds['height']

    return runoff


def runoff(cutout, smooth=None, lower_threshold_quantile=None,
           normalize_using_yearly=None, **params):
    result = cutout.convert_and_aggregate(convert_func=convert_runoff, **params)

    if smooth is not None:
        if smooth is True:
            smooth = 24 * 7
        if "return_capacity" in params.keys():
            result = result[0].rolling(
                time=smooth, min_periods=1).mean(), result[1]
        else:
            result = result.rolling(time=smooth, min_periods=1).mean()

    if lower_threshold_quantile is not None:
        if lower_threshold_quantile is True:
            lower_threshold_quantile = 5e-3
        lower_threshold = pd.Series(result.values.ravel())\
                            .quantile(lower_threshold_quantile)
        result = result.where(result >= lower_threshold, 0.)

    if normalize_using_yearly is not None:
        normalize_using_yearly_i = normalize_using_yearly.index
        if isinstance(normalize_using_yearly_i, pd.DatetimeIndex):
            normalize_using_yearly_i = normalize_using_yearly_i.year
        else:
            normalize_using_yearly_i = normalize_using_yearly_i.astype(int)

        years = (pd.Series(pd.to_datetime(result.coords['time'].values).year)
                 .value_counts().loc[lambda x: x > 8700].index
                 .intersection(normalize_using_yearly_i))
        assert len(years), "Need at least a full year of data (more is better)"
        years_overlap = slice(str(min(years)), str(max(years)))

        dim = result.dims[1 - result.get_axis_num('time')]
        result *= (
            (xr.DataArray(
                normalize_using_yearly.loc[years_overlap].sum(),
                dims=[dim]) /
                result.sel(
                time=years_overlap).sum('time')) .reindex(
                countries=result.coords['countries']))

    return result


def hydro(cutout, plants, hydrobasins, flowspeed=1, weight_with_height=False,
          show_progress=True, **kwargs):
    """
    Compute inflow time-series for `plants` by aggregating over catchment
    basins from `hydrobasins`

    Parameters
    ----------
    plants : pd.DataFrame
        Run-of-river plants or dams with lon, lat columns.
    hydrobasins : str|gpd.GeoDataFrame
        Filename or GeoDataFrame of one level of the HydroBASINS dataset.
    flowspeed : float
        Average speed of water flows to estimate the water travel time from
        basin to plant (default: 1 m/s).
    weight_with_height : bool
        Whether surface runoff should be weighted by potential height (probably
        better for coarser resolution).
    show_progress : bool
        Whether to display progressbars.

    References
    ----------
    [1] Liu, Hailiang, et al. "A validated high-resolution hydro power
    time-series model for energy systems analysis." arXiv preprint
    arXiv:1901.08476 (2019).

    [2] Lehner, B., Grill G. (2013): Global river hydrography and network
    routing: baseline data and new approaches to study the world’s large river
    systems. Hydrological Processes, 27(15): 2171–2186. Data is available at
    www.hydrosheds.org.
    """
    basins = hydrom.determine_basins(
        plants, hydrobasins, show_progress=show_progress)

    matrix = cutout.indicatormatrix(basins.shapes)
    # compute the average surface runoff in each basin
    matrix_normalized = matrix / matrix.sum(axis=1)
    runoff = cutout.runoff(matrix=matrix_normalized, index=basins.shapes.index,
                           weight_with_height=weight_with_height,
                           show_progress=show_progress, **kwargs)
    # The hydrological parameters are in units of "m of water per day" and so
    # they should be multiplied by 1000 and the basin area to convert to m3
    # d-1 = m3 h-1 / 24
    runoff *= (1000. / 24.) * \
        xr.DataArray(basins.shapes.to_crs(dict(proj="cea")).area)

    return hydrom.shift_and_aggregate_runoff_for_plants(
        basins, runoff, flowspeed, show_progress)



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
General utility functions for internal use.
"""

from .gis import maybe_swap_spatial_dims
import progressbar as pgb
from pathlib import Path
import pandas as pd
import xarray as xr
import textwrap
import re
import warnings

from .datasets import modules as datamodules

import logging
logger = logging.getLogger(__name__)


def make_optional_progressbar(show, prefix, max_value=None):
    warnings.warn("make_optional_progressbar() is deprecated and will be removed "
                  "in the next version.", warnings.DeprecationWarning)
    if show:
        widgets = [
            pgb.widgets.Percentage(),
            ' ',
            pgb.widgets.SimpleProgress(
                format='(%s)' %
                pgb.widgets.SimpleProgress.DEFAULT_FORMAT),
            ' ',
            pgb.widgets.Bar(),
            ' ',
            pgb.widgets.Timer(),
            ' ',
            pgb.widgets.ETA()]
        if not prefix.endswith(": "):
            prefix = prefix.strip() + ": "
        maybe_progressbar = pgb.ProgressBar(prefix=prefix, widgets=widgets,
                                            max_value=max_value)
    else:
        def maybe_progressbar(x):
            return x

    return maybe_progressbar


def migrate_from_cutout_directory(old_cutout_dir, path):
    """Convert an old style cutout directory to new style netcdf file"""
    old_cutout_dir = Path(old_cutout_dir)
    with xr.open_dataset(old_cutout_dir / "meta.nc") as meta:
        newname = f"{old_cutout_dir.name}.nc"
        module = meta.attrs["module"]
        minX, maxX = meta.indexes['x'][[0, -1]]
        minY, maxY = sorted(meta.indexes['y'][[0, -1]])
        minT, maxT = meta.indexes['time'][[0, -1]].strftime("%Y-%m")

        logger.warning(textwrap.dedent(f"""
            Found an old-style directory-like cutout. It can manually be
            recreated using

            cutout = atlite.Cutout("{newname}",
                                   module="{module}",
                                   time=slice("{minT}", "{maxT}"),
                                   x=slice({minX}, {maxX}),
                                   y=slice({minY}, {maxY})
            cutout.prepare()

            but we are trying to offer an automated migration as well ...
        """))

        try:
            data = xr.open_mfdataset(str(old_cutout_dir / "[12]*.nc"),
                                     combine="by_coords")
            data.attrs.update(meta.attrs)
        except xr.MergeError:
            logger.exception(
                "Automatic migration failed. Re-create the cutout "
                "with the command above!")
            raise

    data = maybe_swap_spatial_dims(data)
    module = data.attrs["module"]
    data.attrs['prepared_features'] = list(datamodules[module].features)
    for v in data:
        data[v].attrs['module'] = module
        fd = datamodules[module].features.items()
        features = [k for k, l in fd if v in l]
        data[v].attrs['feature'] = features.pop() if features else 'undefined'

    path = Path(path).with_suffix(".nc")
    logger.info(f"Writing cutout data to {path}. When done, load it again using"
                f"\n\n\tatlite.Cutout('{path}')")
    data.to_netcdf(path)
    return data


def timeindex_from_slice(timeslice):
    end = pd.Timestamp(timeslice.end) + pd.offsets.DateOffset(months=1)
    return pd.date_range(timeslice.start, end, freq="1h", closed="left")


class arrowdict(dict):
    """
    A subclass of dict, which allows you to get
    items in the dict using the attribute syntax!
    """

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError as e:
            raise AttributeError(e.args[0])

    _re_pattern = re.compile('[a-zA-Z_][a-zA-Z0-9_]*')

    def __dir__(self):
        dict_keys = []
        for k in self.keys():
            if isinstance(k, str):
                m = self._re_pattern.match(k)
                if m:
                    dict_keys.append(m.string)
        return dict_keys


class CachedAttribute(object):
    '''
    Computes attribute value and caches it in the instance.
    From the Python Cookbook (Denis Otkidach)
    This decorator allows you to create a property which can be
    computed once and accessed many times. Sort of like memoization.
    '''

    def __init__(self, method, name=None, doc=None):
        # record the unbound-method and the name
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = doc or method.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            # instance attribute accessed on class, return self
            # You get here if you write `Foo.bar`
            return self
        # compute, cache and return the instance's attribute value
        result = self.method(inst)
        # setattr redefines the instance's attribute so this doesn't get called
        # again
        setattr(inst, self.name, result)
        return result



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Module containing specific operations for creating cutouts from the NCEP dataset.

DEPRECATED
----------

The ncep dataset module has not been ported to Atlite v0.2, yet. Use Atlite v0.0.2 to use it,
for the time being!
"""

import pandas as pd
import numpy as np
import xarray as xr
import os
import glob
import tempfile
import subprocess
import shutil

engine = 'pynio'
crs = 4326


def convert_lons_lats_ncep(ds, xs, ys):
    if not isinstance(xs, slice):
        first, second, last = np.asarray(xs)[[0, 1, -1]]
        xs = slice(first - 0.1 * (second - first),
                   last + 0.1 * (second - first))
    if not isinstance(ys, slice):
        first, second, last = np.asarray(ys)[[0, 1, -1]]
        ys = slice(first - 0.1 * (second - first),
                   last + 0.1 * (second - first))

    ds = ds.sel(lat_0=ys)

    # Lons should go from -180. to +180.
    if len(
        ds.coords['lon_0'].sel(
            lon_0=slice(
            xs.start +
            360.,
            xs.stop +
            360.))):
        ds = xr.concat([ds.sel(lon_0=slice(xs.start + 360., xs.stop + 360.)),
                        ds.sel(lon_0=xs)],
                       dim="lon_0")
        ds = ds.assign_coords(lon_0=np.where(ds.coords['lon_0'].values <= 180,
                                             ds.coords['lon_0'].values,
                                             ds.coords['lon_0'].values - 360.))
    else:
        ds = ds.sel(lon_0=xs)

    ds = ds.rename({'lon_0': 'x', 'lat_0': 'y'})
    ds = ds.assign_coords(lon=ds.coords['x'], lat=ds.coords['y'])
    return ds


def convert_time_hourly_ncep(ds, drop_time_vars=True):
    # Combine initial_time0 and forecast_time0
    ds = (
        ds .stack(
            time=(
                "initial_time0_hours",
                "forecast_time0")) .assign_coords(
            time=np.ravel(
                ds.coords['initial_time0_hours'] +
                ds.coords['forecast_time0'] -
                pd.Timedelta('1h'))))
    if drop_time_vars:
        ds = ds.drop(['initial_time0', 'initial_time0_encoded'])
    return ds


def convert_unaverage_ncep(ds):
    # the fields ending in _avg contain averages which have to be unaveraged by using
    # \begin{equation}
    # \tilde x_1 = x_1 \quad \tilde x_i = i \cdot x_i - (i - 1) \cdot x_{i-1} \quad \forall i > 1
    # \end{equation}

    def unaverage(da, dim='forecast_time0'):
        coords = da.coords[dim]
        y = da * xr.DataArray(np.arange(1, len(coords) + 1),
                              dims=[dim], coords={dim: coords})
        return y - y.shift(**{dim: 1}).fillna(0.)
    for k, da in ds.items():
        if k.endswith('_avg'):
            ds[k[:-len('_avg')]] = unaverage(da)
            ds = ds.drop(k)

    return ds


def convert_unaccumulate_ncep(ds):
    # the fields ending in _acc contain values that are accumulated over the
    # forecast_time which have to be unaccumulated by using:
    # \begin{equation}
    # \tilde x_1 = x_1
    # \tilde x_i = x_i - x_{i-1} \forall 1 < i <= 6
    # \end{equation}
    # Source:
    # http://rda.ucar.edu/datasets/ds094.1/#docs/FAQs_hrly_timeseries.html

    def unaccumulate(da, dim='forecast_time0'):
        return da - da.shift(**{dim: 1}).fillna(0.)
    for k, da in ds.items():
        if k.endswith('_acc'):
            ds[k[:-len('_acc')]] = unaccumulate(da)
            ds = ds.drop(k)

    return ds


def convert_clip_lower(ds, variable, a_min, value):
    """
    Set values of `variable` that are below `a_min` to `value`.
    Similar to `numpy.clip`.
    """
    ds[variable] = ds[variable].where(ds[variable] > a_min).fillna(value)
    return ds


def prepare_wnd10m_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = convert_time_hourly_ncep(ds)
        ds['wnd10m'] = np.sqrt(
            ds['VGRD_P0_L103_GGA0']**2 +
            ds['UGRD_P0_L103_GGA0']**2)
        ds = ds.drop(['VGRD_P0_L103_GGA0', 'UGRD_P0_L103_GGA0'])
        yield yearmonth, ds


def prepare_influx_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = convert_unaverage_ncep(ds)
        ds = convert_time_hourly_ncep(ds)

        ds = ds.rename({'DSWRF_P8_L1_GGA0': 'influx'})
        # clipping random fluctuations around zero
        ds = convert_clip_lower(ds, 'influx', a_min=0.1, value=0.)
        yield yearmonth, ds


def prepare_outflux_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = convert_unaverage_ncep(ds)
        ds = convert_time_hourly_ncep(ds)

        ds = ds.rename({'USWRF_P8_L1_GGA0': 'outflux'})
        # clipping random fluctuations around zero
        ds = convert_clip_lower(ds, 'outflux', a_min=3., value=0.)
        yield yearmonth, ds


def prepare_temperature_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = convert_time_hourly_ncep(ds)

        ds = ds.rename({'TMP_P0_L103_GGA0': 'temperature'})
        yield yearmonth, ds


def prepare_soil_temperature_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = convert_time_hourly_ncep(ds)

        ds = ds.rename({'TMP_P0_2L106_GGA0': 'soil temperature'})
        yield yearmonth, ds


def prepare_runoff_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        # runoff has missing values: set nans to 0
        ds = ds.fillna(0.)
        ds = convert_unaccumulate_ncep(ds)
        ds = convert_time_hourly_ncep(ds)

        ds = ds.rename({'WATR_P8_L1_GGA0': 'runoff'})
        yield yearmonth, ds


def prepare_height_ncep(fn, xs, ys, yearmonths, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = ds.rename({'HGT_P0_L105_GGA0': 'height'})
        for ym in yearmonths:
            yield ym, ds


def prepare_roughness_ncep(fn, yearmonth, xs, ys, engine=engine):
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = ds.rename({'SFCR_P8_L1_GGA0': 'roughness'})
        ds = ds.assign_coords(
            year=yearmonth[0]).assign_coords(
            month=yearmonth[1])
        yield yearmonth, ds


def prepare_meta_ncep(
        xs,
        ys,
        year,
        month,
        template,
        height_config,
        module,
        engine=engine):
    fn = next(glob.iglob(template.format(year=year, month=month)))
    with xr.open_dataset(fn, engine=engine) as ds:
        ds = ds.coords.to_dataset()
        ds = convert_lons_lats_ncep(ds, xs, ys)
        ds = convert_time_hourly_ncep(ds, drop_time_vars=False)
        meta = ds.load()

    xs = ds['x'].values
    ys = ds['y'].values

    height_config = height_config.copy()
    height_tasks_func = height_config.pop('tasks_func')
    height_task, = height_tasks_func(
        xs, ys, [(year, month)], meta_attrs={}, **height_config)
    height_prepare_func = height_task.pop('prepare_func')
    _, ds = next(height_prepare_func(**height_task))

    meta['height'] = ds['height']

    return meta


def tasks_monthly_ncep(xs, ys, yearmonths, prepare_func, template, meta_attrs):
    return [dict(prepare_func=prepare_func,
                 xs=xs, ys=ys,
                 fn=next(glob.iglob(template.format(year=ym[0], month=ym[1]))),
                 engine=engine,
                 yearmonth=ym)
            for ym in yearmonths]


def tasks_height_ncep(
        xs,
        ys,
        yearmonths,
        prepare_func,
        template,
        meta_attrs,
        **extra_args):
    return [dict(prepare_func=prepare_func,
                 xs=xs, ys=ys, yearmonths=yearmonths,
                 fn=next(glob.iglob(template)),
                 **extra_args)]


weather_data_config = {
    'influx': dict(tasks_func=tasks_monthly_ncep,
                   prepare_func=prepare_influx_ncep,
                   template=os.path.join(config.ncep_dir, '{year}{month:0>2}/dswsfc.*.grb2')),
    'outflux': dict(tasks_func=tasks_monthly_ncep,
                    prepare_func=prepare_outflux_ncep,
                    template=os.path.join(config.ncep_dir, '{year}{month:0>2}/uswsfc.*.grb2')),
    'temperature': dict(tasks_func=tasks_monthly_ncep,
                        prepare_func=prepare_temperature_ncep,
                        template=os.path.join(config.ncep_dir, '{year}{month:0>2}/tmp2m.*.grb2')),
    'soil temperature': dict(tasks_func=tasks_monthly_ncep,
                             prepare_func=prepare_soil_temperature_ncep,
                             template=os.path.join(config.ncep_dir, '{year}{month:0>2}/soilt1.*.grb2')),
    'wnd10m': dict(tasks_func=tasks_monthly_ncep,
                   prepare_func=prepare_wnd10m_ncep,
                   template=os.path.join(config.ncep_dir, '{year}{month:0>2}/wnd10m.*.grb2')),
    'runoff': dict(tasks_func=tasks_monthly_ncep,
                   prepare_func=prepare_runoff_ncep,
                   template=os.path.join(config.ncep_dir, '{year}{month:0>2}/runoff.*.grb2')),
    'roughness': dict(tasks_func=tasks_monthly_ncep,
                      prepare_func=prepare_roughness_ncep,
                      template=os.path.join(config.ncep_dir, '{year}{month:0>2}/flxf.gdas.*.grb2')),
    'height': dict(tasks_func=tasks_height_ncep,
                   prepare_func=prepare_height_ncep,
                   template=os.path.join(config.ncep_dir, 'height/cdas1.20130101.splgrbanl.grb2'))
}

meta_data_config = dict(
    prepare_func=prepare_meta_ncep,
    template=os.path.join(
        config.ncep_dir,
        '{year}{month:0>2}/tmp2m.*.grb2'),
    height_config=weather_data_config['height'])



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import era5, sarah, gebco

modules  = {'era5': era5, 'sarah': sarah, 'gebco': gebco}


# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

from setuptools import setup, find_packages
from codecs import open

with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='atlite',
    author='The Atlite Authors',
    author_email='jonas.hoersch@posteo.de',
    description='Library for fetching and converting weather data to power systems data',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/PyPSA/atlite',
    license='GPLv3',
    packages=find_packages(exclude=['doc', 'test']),
    include_package_data=True,
    python_requires='~=3.6',
    use_scm_version={'write_to': 'atlite/version.py'},
    setup_requires=['setuptools_scm'],
    install_requires=['numpy',
                      'scipy',
                      'pandas>=0.22',
                      'bottleneck',
                      'numexpr',
                      'xarray>=0.16',
                      'netcdf4',
                      'dask>=0.18.0, <20.04.0',
                      'requests',
                      'rtree',
                      "pyyaml",
                      'rasterio>=1.0',
                      'shapely',
                      'progressbar2',
                      'pyproj>=2',
                      'geopandas',
                      'cdsapi'],
    extras_require = {
        "docs": ["numpydoc", "sphinx", "sphinx_rtd_theme",
		 "nbsphinx", "nbsphinx-link"]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ])



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Atlite helps you to convert weather data into energy systems model data.

Atlite is a free software, xarray-based Python library for converting weather data
(like wind speeds) into energy systems data. It is designed to by lightweight and
work with big weather datasets while keeping the resource requirements especially
on CPU and RAM resources low.
"""

from .cutout import Cutout
from .gis import compute_indicatormatrix, regrid, ExclusionContainer
from .resource import windturbines, solarpanels

from .version import version as __version__

__author__ = ("The Atlite Authors: Gorm Andresen (Aarhus University), "
              "Jonas Hoersch (FIAS/KIT/RLI), "
              "Johannes Hampp (JLUG),"
              "Fabian Hofmann (FIAS)"
              "Tom Brown (FIAS/KIT), "
              "Markus Schlott (FIAS), "
              "David Schlachtberger (FIAS), "
              )
__copyright__ = "Copyright 2016-2021 The Atlite Authors"



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

from numpy import pi
import xarray as xr
from numpy import sin, cos, arcsin, arccos, arctan2, deg2rad


def SolarPosition(ds):
    """
    Compute solar azimuth and altitude

    Solar altitude errors are up to 1.5 deg during sun-rise and set, but at
    0.05-0.1 deg during daytime.

    References
    ----------
    [1] Michalsky, J. J., The astronomical almanac’s algorithm for approximate
    solar position (1950–2050), Solar Energy, 40(3), 227–235 (1988).
    [2] Sproul, A. B., Derivation of the solar geometric relationships using
    vector analysis, Renewable Energy, 32(7), 1187–1205 (2007).
    [3] Kalogirou, Solar Energy Engineering (2009).

    More accurate algorithms would be
    ---------------------------------
    [4] I. Reda and A. Andreas, Solar position algorithm for solar
    radiation applications. Solar Energy, vol. 76, no. 5, pp. 577-589, 2004.
    [5] I. Reda and A. Andreas, Corrigendum to Solar position algorithm for
    solar radiation applications. Solar Energy, vol. 81, no. 6, p. 838, 2007.
    [6] Blanc, P., & Wald, L., The SG2 algorithm for a fast and accurate
    computation of the position of the sun for multi-decadal time period, Solar
    Energy, 86(10), 3072–3083 (2012).

    The unfortunately quite computationally intensive SPA algorithm [4,5] has
    been implemented using numba or plain numpy for a single location at
    https://github.com/pvlib/pvlib-python/blob/master/pvlib/spa.py.

    """

    # up to h and dec from [1]

    t = ds.indexes['time']
    n = xr.DataArray(t.to_julian_date(), [t]) - 2451545.0
    hour = ds['time.hour']
    minute = ds['time.minute']

    if 'time' in ds.chunks:
        chunks = {'time': ds.chunks['time']}
        n = n.chunk(chunks)
        hour = hour.chunk(chunks)
        minute = minute.chunk(chunks)

    L = 280.460 + 0.9856474 * n  # mean longitude (deg)
    g = deg2rad(357.528 + 0.9856003 * n)  # mean anomaly (rad)
    l = deg2rad(
        L +
        1.915 *
        sin(g) +
        0.020 *
        sin(2 * g))  # ecliptic long. (rad)
    ep = deg2rad(23.439 - 4e-7 * n)  # obliquity of the ecliptic (rad)

    ra = arctan2(cos(ep) * sin(l), cos(l))  # right ascencion (rad)
    lmst = (6.697375 + (hour + minute / 60.0) +
            0.0657098242 * n) * 15. + ds['lon']  # local mean sidereal time (deg)
    h = (deg2rad(lmst) - ra + pi) % (2 * pi) - \
        pi  # hour angle (rad)

    dec = arcsin(sin(ep) * sin(l))            # declination (rad)

    # alt and az from [2]
    lat = deg2rad(ds['lat'])
    # Clip before arcsin to prevent values < -1. from rounding errors; can
    # cause NaNs later
    alt = arcsin(
        (sin(lat) *
          sin(dec) +
          cos(lat) *
          cos(dec) *
          cos(h)).clip(min=-1., max=1.)).rename('altitude')

    az = arccos(
        ((sin(dec) *
          cos(lat) -
          cos(dec) *
          sin(lat) *
          cos(h)) /
          cos(alt)).clip(min=-1., max=1.))
    az = az.where(h <= 0, 2 * pi - az).rename('azimuth')

    if 'influx_toa' in ds:
        atmospheric_insolation = ds['influx_toa'].rename(
            'atmospheric insolation')
    else:
        # [3]
        atmospheric_insolation = (1366.1 * (1 + 0.033 * cos(g)) * sin(alt))\
                                  .rename('atmospheric insolation')

    vars = {da.name: da for da in [alt, az, atmospheric_insolation]}
    solar_position = xr.Dataset(vars)
    return solar_position



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import pandas as pd
import xarray as xr
from numpy import sin, cos, fmin, fmax, sqrt, deg2rad
import logging
logger = logging.getLogger(__name__)


def DiffuseHorizontalIrrad(ds, solar_position, clearsky_model, influx):
    # Clearsky model from Reindl 1990 to split downward radiation into direct
    # and diffuse contributions. Should switch to more up-to-date model, f.ex.
    # Ridley et al. (2010) http://dx.doi.org/10.1016/j.renene.2009.07.018 ,
    # Lauret et al. (2013):http://dx.doi.org/10.1016/j.renene.2012.01.049

    sinaltitude = sin(solar_position['altitude'])
    atmospheric_insolation = solar_position['atmospheric insolation']

    if clearsky_model is None:
        clearsky_model = ('enhanced'
                          if 'temperature' in ds and 'humidity' in ds
                          else 'simple')

    # Reindl 1990 clearsky model

    k = influx / atmospheric_insolation  # clearsky index
    # k.values[k.values > 1.0] = 1.0
    # k = k.rename('clearsky index')

    if clearsky_model == 'simple':
        # Simple Reindl model without ambient air temperature and
        # relative humidity
        fraction = (((k > 0.0) & (k <= 0.3)) *
                    fmin(1.0, 1.020 - 0.254 * k + 0.0123 * sinaltitude) +
                    ((k > 0.3) & (k < 0.78)) *
                    fmin(0.97, fmax(0.1, 1.400 - 1.749 * k + 0.177 *
                                          sinaltitude)) +
                    (k >= 0.78) *
                    fmax(0.1, 0.486 * k - 0.182 * sinaltitude))
    elif clearsky_model == 'enhanced':
        # Enhanced Reindl model with ambient air temperature and relative
        # humidity
        T = ds['temperature']
        rh = ds['humidity']

        fraction = (((k > 0.0) & (k <= 0.3)) *
                    fmin(1.0, 1.000 - 0.232 * k + 0.0239 * sinaltitude -
                            0.000682 * T + 0.0195 * rh) +
                    ((k > 0.3) & (k < 0.78)) *
                    fmin(0.97, fmax(0.1, 1.329 - 1.716 * k + 0.267 *
                                          sinaltitude - 0.00357 * T +
                                          0.106 * rh)) +
                    (k >= 0.78) *
                    fmax(0.1, 0.426 * k - 0.256 * sinaltitude + 0.00349 *
                            T + 0.0734 * rh))
    else:
        raise KeyError("`clearsky model` must be chosen from 'simple' and "
                       "'enhanced'")

    # Set diffuse fraction to one when the sun isn't up
    # fraction = fraction.where(sinaltitude >= sin(deg2rad(threshold))).fillna(1.0)
    # fraction = fraction.rename('fraction index')

    return (influx * fraction).rename('diffuse horizontal')


def TiltedDiffuseIrrad(
        ds,
        solar_position,
        surface_orientation,
        direct,
        diffuse):
    # Hay-Davies Model

    sinaltitude = sin(solar_position['altitude'])
    atmospheric_insolation = solar_position['atmospheric insolation']

    cosincidence = surface_orientation['cosincidence']
    surface_slope = surface_orientation['slope']

    influx = direct + diffuse

    with np.errstate(divide='ignore', invalid='ignore'):
        # brightening factor
        f = sqrt(direct / influx).fillna(0.)

        # anisotropy factor
        A = direct / atmospheric_insolation

    # geometric factor
    R_b = cosincidence / sinaltitude

    diffuse_t = ((1.0 - A) * ((1 + cos(surface_slope)) / 2.0) *
                 (1.0 + f * sin(surface_slope / 2.0)**3)
                 + A * R_b) * diffuse

    # fixup: clip all negative values (unclear why it gets negative)
    # note: REatlas does not do the fixup
    if logger.isEnabledFor(logging.WARNING):
        if ((diffuse_t < 0.) & (sinaltitude > sin(deg2rad(1.)))).any():
            logger.warning(
                'diffuse_t exhibits negative values above altitude threshold.')

    with np.errstate(invalid='ignore'):
        diffuse_t = diffuse_t.clip(min=0).fillna(0)

    return diffuse_t.rename('diffuse tilted')


def TiltedDirectIrrad(solar_position, surface_orientation, direct):
    sinaltitude = sin(solar_position['altitude'])
    cosincidence = surface_orientation['cosincidence']

    # geometric factor
    R_b = cosincidence / sinaltitude

    return (R_b * direct).rename('direct tilted')


def _albedo(ds, influx):
    if 'albedo' in ds:
        albedo = ds['albedo']
    elif 'outflux' in ds:
        albedo = (ds['outflux'] / influx.where(influx!=0)).fillna(0).clip(max=1)
    else:
        raise AssertionError(
            "Need either albedo or outflux as a variable in the dataset. "
            "Check your cutout and dataset module.")

    return albedo


def TiltedGroundIrrad(ds, solar_position, surface_orientation, influx):
    surface_slope = surface_orientation['slope']
    ground_t = influx * _albedo(ds, influx) * \
        (1.0 - cos(surface_slope)) / 2.0
    return ground_t.rename('ground tilted')


def TiltedIrradiation(ds, solar_position, surface_orientation, trigon_model,
                      clearsky_model, altitude_threshold=1.):

    influx_toa = solar_position['atmospheric insolation']

    def clip(influx, influx_max):
        # use .data in clip due to dask-xarray incompatibilities
        return influx.clip(min=0, max=influx_max.transpose(*influx.dims).data)

    if 'influx' in ds:
        influx = clip(ds['influx'], influx_toa)
        diffuse = DiffuseHorizontalIrrad(
            ds, solar_position, clearsky_model, influx)
        direct = influx - diffuse
    elif 'influx_direct' in ds and 'influx_diffuse' in ds:
        direct = clip(ds['influx_direct'], influx_toa)
        diffuse = clip(ds['influx_diffuse'], influx_toa - direct)
    else:
        raise AssertionError(
            "Need either influx or influx_direct and influx_diffuse in the "
            "dataset. Check your cutout and dataset module.")
    if trigon_model == 'simple':
        k = surface_orientation['cosincidence'] / \
            sin(solar_position['altitude'])
        cos_surface_slope = cos(surface_orientation['slope'])

        influx = direct + diffuse
        direct_t = k * direct
        diffuse_t = ((1. + cos_surface_slope) / 2. * diffuse +
                     _albedo(ds, influx) * influx * ((1. - cos_surface_slope) / 2.))

        total_t = direct_t.fillna(0.) + diffuse_t.fillna(0.)
    else:
        diffuse_t = TiltedDiffuseIrrad(
            ds, solar_position, surface_orientation, direct, diffuse)
        direct_t = TiltedDirectIrrad(
            solar_position, surface_orientation, direct)
        ground_t = TiltedGroundIrrad(
            ds,
            solar_position,
            surface_orientation,
            direct + diffuse)

        total_t = (direct_t + diffuse_t + ground_t).rename('total tilted')

    # The solar_position algorithms have a high error for small solar altitude
    # values, leading to big overall errors from the 1/sinaltitude factor.
    # => Suppress irradiation below solar altitudes of 1 deg.

    cap_alt = solar_position['altitude'] < deg2rad(altitude_threshold)
    total_t = total_t.where(~(cap_alt | (direct + diffuse <= 0.01)), 0)

    return total_t
